unit uMain;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Dialogs, StdCtrls, ExtCtrls, Graphics,
  Menus, LCLIntf, INIFiles,

  DateUtils, sntpsend;

type

  TMyNtpData = packed record
    lastattempt: String;
    lastsync: String;
    connected: String;
  end;

type

  TNtpData = packed record
    mode: Byte;
    stratum: Byte;
    poll: Byte;
    Precision: Byte;
    RootDelay: Longint;
    RootDisperson: Longint;
    RefID: Longint;
    Ref1: Longint;
    Ref2: Longint;
    Org1: Longint;
    Org2: Longint;
    Rcv1: Longint;
    Rcv2: Longint;
    Xmit1: Longint;
    Xmit2: Longint;
  end;


type

  { TMainForm }

  TMainForm = class(TForm)
    lblModeDesc: TLabel;
    lblOrg2Desc: TLabel;
    lblRcv1Desc: TLabel;
    lblRcv2Desc: TLabel;
    lblXmit1Desc: TLabel;
    lblXmit2Desc: TLabel;
    lblMode: TLabel;
    lblPoll: TLabel;
    lblPrecision: TLabel;
    lblRootDelay: TLabel;
    lblRootDispersion: TLabel;
    lblPollDesc: TLabel;
    lblRefID: TLabel;
    lblRef1: TLabel;
    lblRef2: TLabel;
    lblOrg1: TLabel;
    lblOrg2: TLabel;
    lblRcv1: TLabel;
    lblRcv2: TLabel;
    lblXmit1: TLabel;
    lblXmit2: TLabel;
    lblPrecisionDesc: TLabel;
    lblRootDelayDesc: TLabel;
    lblRootDispersionDesc: TLabel;
    lblRefIDDesc: TLabel;
    lblRef1Desc: TLabel;
    lblRef2Desc: TLabel;
    lblOrg1Desc: TLabel;
    lblStratum: TLabel;
    UpdateTime: TButton;
    SaveSettings: TButton;
    MaxSyncDiffUnit: TComboBox;
    UpdateIntervalUnit: TComboBox;
    MaxSyncDiff: TEdit;
    TimeServer: TEdit;
    UpdateInterval: TEdit;
    gbxTimeInfo: TGroupBox;
    gbxNTPInfo: TGroupBox;
    gbxSettings: TGroupBox;
    lblTimeDesc: TLabel;
    lblMaxSyncDiff: TLabel;
    lblTime: TLabel;
    lblUTC: TLabel;
    lblLastAttempt: TLabel;
    lblNextAttempt: TLabel;
    lblLastSync: TLabel;
    lblDiffDesc: TLabel;
    lblDiff: TLabel;
    lblUTCDesc: TLabel;
    lblLastAttemptDesc: TLabel;
    lblLastSyncDesc: TLabel;
    lblNextAttemptDesc: TLabel;
    lblStratumDesc: TLabel;
    lblTimeServer: TLabel;
    lblUpdateInterval: TLabel;
    RealTimer: TTimer;
    UpdateIntervalTimer: TTimer;

    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure FormShow(Sender: TObject);
    procedure SaveSettingsClick(Sender: TObject);
    procedure UpdateTimeClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure RealTimerTimer(Sender: TObject);
    procedure UpdateIntervalTimerTimer(Sender: TObject);

    //
    procedure gbxTimeData;
    procedure gbxServerData;
    procedure UpdateNTP;
    procedure SetUpdateInterval;
    procedure CheckInterval(aInterval: integer);
    procedure SaveIniFile;
    procedure LoadIniFile;
  private

  public
    FInterval: integer;
    ini: TIniFile;
  end;

var
  MainForm: TMainForm;
  MyNtpData: TMyNtpData;
  NtpData: TNtpData;

implementation

{$R *.lfm}

{ TMainForm }

procedure TMainForm.FormCreate(Sender: TObject);
begin
  LoadIniFile;

  Screen.HintFont.Name := 'Courier New';
  Screen.HintFont.Size := 8;

  // 1. Disable interval timer
  UpdateIntervalTimer.Enabled:=false;
  // 2. Get data from ntp server
  UpdateNTP;
  // 3. Show data in form
  gbxTimeData;
  gbxServerData;

  // include hints for time server informations
  {$i uhints.pas}
end;

// Show date and time in form
procedure TMainForm.gbxTimeData;
begin
  lblTime.Caption := DateTimeToStr(now);
  lblUTC.Caption := DateTimeToStr(LocalTimeToUniversal(now));
  lblDiff.Caption := Format('%dh', [GetLocalTimeOffset div 60]);
  lblLastAttempt.Caption := MyNtpData.lastattempt;
  lblLastSync.Caption := MyNtpData.lastsync;
end;

{ Berechnet die Differenz zw. "now" und einem Timestamp in ms.
Die Ausgabe erfolgt als String im DateTime Format }
function DTDiffToStr(aDT1: TDateTime; aDT2: Longint): String;
begin
  Result := DateTimeToStr(TimeStampToDateTime(MSecsToTimeStamp(
            TimeStampToMSecs(DateTimeToTimeStamp(aDT1)) + aDT2)));
end;

// Show ntp server related data
procedure TMainForm.gbxServerData;
begin
  lblMode.Caption := NtpData.mode.ToString;
  lblStratum.Caption := NtpData.stratum.ToString;
  lblPoll.Caption := NtpData.poll.ToString;
  lblPrecision.Caption := NtpData.Precision.ToString;
  lblRootDelay.Caption := NtpData.RootDelay.ToString;
  lblRootDispersion.Caption := NtpData.RootDisperson.ToString;
  lblRefID.Caption := NtpData.RefID.ToString;
  lblRef1.Caption := NtpData.Ref1.ToString;
  lblRef2.Caption := NtpData.Ref2.ToString;
  lblOrg1.Caption := NtpData.Org1.ToString;
  lblOrg2.Caption := NtpData.Org2.ToString;
  lblRcv1.Caption := NtpData.Rcv1.ToString;
  lblRcv2.Caption := NtpData.Rcv2.ToString;
  lblXmit1.Caption := NtpData.Xmit1.ToString;
  lblXmit2.Caption := NtpData.Xmit2.ToString;
end;

// manual update
procedure TMainForm.UpdateTimeClick(Sender: TObject);
begin
  UpdateNTP;
end;

procedure TMainForm.SaveSettingsClick(Sender: TObject);
begin
  SaveIniFile;
end;

procedure TMainForm.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
  SaveIniFile;
end;

procedure TMainForm.FormShow(Sender: TObject);
begin
  UpdateTime.SetFocus;
end;

// Update date and time in form
procedure TMainForm.RealTimerTimer(Sender: TObject);
begin
  gbxTimeData;
end;

// Countdown timer for next auto update
procedure TMainForm.UpdateIntervalTimerTimer(Sender: TObject);
begin
  lblNextAttempt.Caption := Format('%.2dh %.2dm %.2ds', [FInterval div 3600, FInterval div 60 mod 60, FInterval mod 60]);
  Dec(FInterval);
  if FInterval < 0 then begin
    UpdateIntervalTimer.Enabled:=false;
    UpdateNTP;
  end;
end;

// Get time information from ntp server
procedure TMainForm.UpdateNTP;
var
  sntp: TSNTPSend;
begin
  SetUpdateInterval;

  sntp := TSNTPSend.Create;
  sntp.Timeout:=2000;
  sntp.TargetHost:='de.pool.ntp.org';
  sntp.SyncTime:=true;
  sntp.MaxSyncDiff:=86400; // 1 day
  try
    if sntp.GetNTP or sntp.GetSNTP then
      begin
        MyNtpData.lastattempt := DateTimeToStr(now);
        MyNtpData.lastsync := DateTimeToStr(now) + ' ' + FormatFloat('0',sntp.NTPOffSet*1000) + 'ms';
        MyNtpData.connected := 'Contacted';

        // copy record a to b
        Move(sntp.NTPReply, NtpData, SizeOf(NtpData));
      end
      else MyNtpData.connected := 'Not contacted!';

  finally
    sntp.free;
  end;

  gbxServerData;
end;

// Convert update interval value to seconds
procedure TMainForm.SetUpdateInterval;
begin
  // disable update timer if running
  if UpdateIntervalTimer.Enabled then UpdateIntervalTimer.Enabled:=false;

  // convert interval string to integer
  TryStrToInt(Trim(UpdateInterval.Text), FInterval);

  // convert input to seconds
  case UpdateIntervalUnit.ItemIndex of
    0: FInterval := FInterval; // seconds
    1: FInterval *= 60;    // minutes
    2: FInterval *= 3600;  // hours
    3: FInterval *= 86400; // days
  end;

  // check interval
  CheckInterval(FInterval);

  // enable update timer
  UpdateIntervalTimer.Enabled:=true;
end;

// Check the interval value for validity
procedure TMainForm.CheckInterval(aInterval: integer);
begin
  // the minimum update interval for ntp servers pool.ntp.org is 15 minutes
  if aInterval < 900 then begin
    MessageDlg('Attention!','The minimum update interval for ntp servers pool.ntp.org is 15 minutes!', mtInformation ,[mbOK],0);
    FInterval := 900;
    UpdateInterval.Text:='15';
    UpdateIntervalUnit.ItemIndex := 1;
  end;
end;

procedure TMainForm.SaveIniFile;
begin
  ini := TIniFile.Create(ExtractFilePath(ParamStr(0)) + 'ntpclient.ini');
  try
    try
      ini.WriteString('ntpclient','TimeServer', TimeServer.Text);
      ini.WriteString('ntpclient','UpdateInterval', UpdateInterval.Text);
      ini.WriteInteger('ntpclient','UpdateIntervalUnit', UpdateIntervalUnit.ItemIndex);
      ini.WriteString('ntpclient','MaxSyncDiff', MaxSyncDiff.Text);
      ini.WriteInteger('ntpclient','MaxSyncDiffUnit', MaxSyncDiffUnit.ItemIndex);
    except
      ShowMessage('Error saving ini file!');
    end;
  finally
    ini.Free;
  end;
end;

procedure TMainForm.LoadIniFile;
begin
  ini := TIniFile.Create(ExtractFilePath(ParamStr(0)) + 'ntpclient.ini');
  try
    try
      TimeServer.Text := ini.ReadString('ntpclient','TimeServer', 'de.pool.ntp.org');
      UpdateInterval.Text := ini.ReadString('ntpclient','UpdateInterval', '12');
      UpdateIntervalUnit.ItemIndex := ini.ReadInteger('ntpclient','UpdateIntervalUnit', 2);
      MaxSyncDiff.Text := ini.ReadString('ntpclient','MaxSyncDiff', '24');
      MaxSyncDiffUnit.ItemIndex := ini.ReadInteger('ntpclient','MaxSyncDiffUnit', 2);
    except
      ShowMessage('Error reading ini file!');
    end;
  finally
    ini.Free;
  end;
end;


end.

